/*
 * Example of client using TCP protocol.  *** Teenie Telnet ***
 */

#include	<pwd.h>
#include	"tcp.h"

extern char *getlogin();
extern struct passwd *getpwuid();

#define	ESC_CHAR	('X'^64)	/* Ctrl-X */

#ifndef INADDR_NONE
#define INADDR_NONE	0xffffffff
#endif

#define TIME_OUT 	60

#define ESC	27		/* ^[  The escape key */
#define DEL	127		/* ^?  The delete key */

extern int t_errno;
int fd=(-1), string=0, ascii=0;
int makeraw=0, raw=0, telnet=0;
int port, timed_out=0;
int flagsuspend=0, connector=0;
int netdebug=0;	

void 	doit(), die();
int	towrite();

#ifdef SIGTSTP
int	background();
#endif

main(argc, argv)
int	argc;
char	*argv[];
{
	extern int optind;

	int c;
	static char Usage[]={"Usage: bgtcli [-r|-s] [-t] [-a] address [port]\n"};

	struct sockaddr_in	serv_addr;
	struct hostent		*hp;

	pname = argv[0];

	/* Option checking. :) */

	while ( (c=getopt(argc, argv, "xarst")) != EOF )
	{
		switch (c)
		{
			case 'r': makeraw=1;
				  break;
			case 's': string=1;
				  break;
			case 't': telnet=1;
				  break;
			case 'a': ascii=1;
				  break;
			case 'x': netdebug=1;
				  break;
			case '?': fprintf(stderr, "%s", Usage);
				  exit(1);
				  break;
			default:  fprintf(stderr, "%s", Usage);
		}
	}

	/* repoint argv */ 
	argc-=optind;
	argv[1]=argv[optind];
	argv[2]=argv[optind+1];
#ifdef OLDDEBUG
	fprintf(stderr, "argv[optind]=%s\n", argv[optind]);
	fprintf(stderr, "argc=%d\n", argc);
#endif

	/* Error check.. */
	if ( makeraw && string )
	{
		fprintf(stderr, "Can't specify both \'-r\' and \'-s\' options.\n");
		fprintf(stderr, "%s", Usage);
		exit(1);
	}

	if ( argc > 2 || argc < 1 )
	{
		fprintf(stderr, "%s", Usage);
		exit(1);
	}

	if ( argc == 2 )
	{
		if ( (port=atoi(argv[2])) <= 0 )
		{
			fprintf(stderr, "tcli: port must be an integer\r\n");
			fprintf(stderr, "%s", Usage);
			exit(1);
		}
	}
	else
	{
		port=23;  /* Default telnet port */
		telnet=1;
	}

#ifdef OLDDEBUG
	fprintf(stderr, "port: %d\n", port);
#endif

	/*
	 * Fill in the structure "serv_addr" with the address of the
	 * server that we want to connect with.
	 */


	/* empty our structure. :) */
	d_zero((char *) &serv_addr, sizeof(serv_addr));

	
	if ( (serv_addr.sin_addr.s_addr=inet_addr(argv[1])) == INADDR_NONE )
	{  /* It's not a dotted-decimal address */

		if ( (hp=gethostbyname(argv[1])) == NULL )
		{
#ifdef HOST_NOT_FOUND
			m_herror(argv[1]);
#else
			fprintf(stderr, "%s: host name error.\n", argv[1]);
#endif
			exit(1);
		}
		else 
			d_copy(hp->h_addr, (char *)&serv_addr.sin_addr, hp->h_length);
	}
	serv_addr.sin_family      = AF_INET;
	serv_addr.sin_port        = htons(port);

	/*
	 * Open a TCP socket (an Internet stream socket).
	 */

	if ( (fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		err_sys("client: can't open stream socket");

	/*
	 * Connect to the server.
	 */

	printf("Connecting to address: %s %d\r\n",
		inet_ntoa(serv_addr.sin_addr), port);

	/* Before we get started, set up signal handling */
	signal(SIGINT, die);
	signal(SIGQUIT, die);
	signal(SIGALRM, die);
	signal(SIGTERM, die);

#ifdef SIGTSTP
	if ( signal(SIGTSTP, SIG_IGN) != SIG_DFL )
	{
		signal(SIGCONT, SIG_IGN);
		flagsuspend=0;
	}
	else
	{
		signal(SIGTSTP, SIG_DFL);
		flagsuspend=1;
	}
#endif /* SIGTSTP */

	/* Set up timeout just in case */
	alarm(TIME_OUT);

	timed_out=1;
	if (connect(fd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
	{
		timed_out=0;
		err_ret("can't connect to server");
		die();
	}

	/* We connected, disable the alarm and then doit! :) */

	printf("Connected.\r\n");
	timed_out=0; alarm(0);
	doit(fd);	/* do it all */

	/*NOTREACHED*/
}

/*
 * Read from stdin, writing to the transport endpoint
 * (to the server process), while reading concurrently from
 * the transport endpoint and printing on the standard output.
 */

void doit(fd)
int	fd;
{
	int	i, n, flags;
	char	line[BUFSIZ];
	fd_set	readfds, writefds;
	int	ttyfd=0;
	struct Buf_Len received, *newbuflen;


	/* Zero out fdset for select() */
	FD_ZERO(&readfds); 

	/* We are connected, now is the best time to stty raw */
	if ( makeraw )
	{
		(void) tty_getmode(0);
		(void) tty_raw(0);
		raw=1;
	}

	if ( telnet )
		init_telnet(fd, ( raw ? 0 : 1 ));

	/*
 	 * Now do that I/O thang. :)
 	 */

	for ( ; ; )
	{
		FD_SET(fd, &readfds);  FD_SET(ttyfd, &readfds);
		
		select(10, &readfds, 0, 0, NULL);

		if ( FD_ISSET(fd, &readfds) )
		{ 
			if ( (received.len=read(fd, received.buffer, BUFSIZ)) 
									<= 0 )
			{
#ifdef OLDDEBUG
				perror("Read error from network");
				fprintf(stderr, "\r");
#endif
				die();
			}

#ifdef BIGDEBUG
			fprintf(stderr, "received.len: %d\r\n", received.len);
#endif
			if ( telnet )
				newbuflen=negotiate(fd, &received);
			else
				newbuflen=(&received);
#ifdef BIGDEBUG
			fprintf(stderr, "newbuflen->len: %d\r\n", 
							newbuflen->len);
#endif

	/* Be aware that according to the telnet protocol, a CR by
	   itself will be sent as a CR-NUL sequence, and a CR-LF 
	   will be sent as a CR-LF sequence.  I have ignored this,
	   in the recieving side, because most terminals filter
	   out NUL characters as padding characters.  To filter
	   out the NULs, define MAPOUTNUL.  */

#define MAPOUTNUL
#ifdef MAPOUTNUL
			if ( telnet )
			{  /* Map cr-null to cr */
				int mapped=0, at_cr=0, j;

				for ( i=0, j=0; i<newbuflen->len; ++i )
				{  
					if ( at_cr )
					{ 
						at_cr=0;
						switch (newbuflen->buffer[i])
						{
							case '\0':
#ifdef OLDDEBUG
						fprintf(stderr, "MAP\r\n");
#endif
							mapped=1;
							continue;
							default:  break;
						}
					}
					if ( newbuflen->buffer[i] == '\r' )
						at_cr=1;
					line[j++]=newbuflen->buffer[i];
				}

				/* Only bother re-copying if we mapped a NUL */
				if ( mapped )
				{
					strncpy(newbuflen->buffer, line, j);
					newbuflen->len=j;
					mapped=0;
				}
			}
#endif /* MAPOUTNUL */
			if ( ascii )
				towrite(1, newbuflen->buffer, newbuflen->len);
			else
				writen(1, newbuflen->buffer, newbuflen->len);
		}

		if ( FD_ISSET(ttyfd, &readfds) )
		{
			if ( (n=read(ttyfd, line, BUFSIZ)) <= 0 )
			{
#ifdef OLDDEBUG
				perror("Read error from tty");
				fprintf(stderr, "\r");
#endif
				die();
			}

			if ( line[0] == ESC_CHAR )
			{
				if ( n == 1 )
					read(ttyfd, (&line[1]), 1);
	
				switch (line[1])
				{
					case 'x': /* Quit */
						  die();
					 	  break;
					case 'z': if ( flagsuspend )
				          	  {  
#ifdef SIGTSTP
							if ( raw )
							{
							    write(1, "\r\n", 2);
							    tty_reset(0);
							}
							kill(0, SIGTSTP);

							if ( raw )
								tty_raw(0);
#else
							fprintf(stderr, 
				"\r\nSystem doesn't support job control.\r\n");
#endif  /* SIGTSTP */			

					  	  }
					  	  else
					  	  {
							fprintf(stderr, 
				"\r\nWaiting shell does not support job control.\r\n");
					  	  }
#ifdef OLDDEBUG
					  	  fprintf(stderr, "Reawakened!\r\n");
#endif
					  	  continue;
					  	  break;
#ifdef SIGTSTP
					case 's': if ((ttyfd=background()) < 0)
							die();
#else
					case 's': fprintf(stderr, 
				"\r\nThis system doesn't support backgrounding.\r\n");
#endif
						  continue;
						  break;
					default:  break;
				}  /* end switch */
			}
						
			if ( string )
			{  /* Convert from line format to string format */
				if ( line[n-1] == '\n' )
					line[n-1]='\0';
			}
			if ( raw && telnet )
			{  /* Convert cr to cr-lf */
				if ( line[n-1] == '\r' )
					line[n++]='\n';
			}
			if (writen(fd, line, n) != n)
			{
				perror("socket write error");
				die();
			}
		}
	}
	/* NOTREACHED */
}


/* A function to write only printable characters from a buffer to a fildes */

towrite(fd, buf, len)
int fd;
char buf[];
int len;
{
	int i, j=0;
	char writebuf[BUFSIZ];

	for ( i=0; i<len; ++i )
	{  /* Any special characters I've missed? :) */

		if (((buf[i] <= DEL) && (buf[i] >= ' ')) || (buf[i] == '\n') ||
		     (buf[i] == '\t') || (buf[i] == '\r') || (buf[i] == ESC))
		{
			writebuf[j++]=buf[i];
		}
		if ( j == BUFSIZ )
		{
			writen(fd, writebuf, j);
			j=0;
		}
	}
	writen(fd, writebuf, j);
	return(j);
}



/* Well, we're done. */

void die(sig)
int sig;
{
	if ( timed_out )
		fprintf(stderr, "Connection timed out.\r\n");


	if ( raw )
	{
#ifdef OLDDEBUG
fprintf(stderr, "Raw..\n");
#endif
		write(1, "\r", 1);
		tty_reset(0);
	}

	printf("Disconnected.\n");

	if ( connector )
		kill(connector, SIGTERM);

	exit(0);
}

void ignore()
{
	;
}


#ifdef SIGTSTP
int background()
{
	int i, pid, tty;
	char tty_name[12], holdfile[18], sigfile[18];
	char user[14], *login;
	struct passwd *pw;
	struct stat statbuff;
	FILE *info;

	/* Check to see if there is already a process waiting */
	if ( (login=getlogin()) == NULL )
	{
		if ( (pw=getpwuid(getuid())) == NULL )
		{
			fprintf(stderr, "Who are you??\r\n");
			return(-1);
		}
		else
			strncpy(user, pw->pw_name, 14);
	}
	else
		strncpy(user, login, 14);
	sprintf(holdfile, "/tmp/%s.fg", user);
	if ( stat(holdfile, &statbuff) == 0 )
	{
		fprintf(stderr, "A background process already exists.\r\n");
		return(0);
	}
	
	if ( (pid=fork()) < 0 )
	{
		perror("fork");
		die();
	}
	else if ( pid == 0 )
	{
		/* set it up... */
		for ( i=0; i<=2; ++i )
		{
			if (isatty(i))
				close(i);
		}
		setpgrp();
	
		signal(SIGUSR2, ignore);

		/* Set up our process id to be connected to */
		if ( (info=fopen(holdfile, "w")) == NULL )
			return(-1);
		fprintf(info, "%d %d\n", getpid(), raw);
		fclose(info);
wait:
		pause();  /* wait for a signal */
	
		sprintf(sigfile, "/tmp/%s.sig", user);
		if ( (info=fopen(sigfile, "r")) == NULL )
			goto wait;
		
		fscanf(info, "%s", tty_name);
		fscanf(info, "%d", &connector);
		unlink("/tmp/signal");
	
		if ( (tty=open(tty_name, O_RDWR)) >= 0 )
			write(tty, "Connected.\r\n", 12);

		/* Fill out stdin, stdout, stderr */
		dup2(tty, 0); dup2(tty, 1); dup2(tty, 2);

		/* We need to change to the process group of the
		   terminal so we can do i/o without receiving EIO */

		setpgrp(getpid(), getpgrp(connector));
	
#ifdef OLDDEBUG
		fprintf(stderr, "tty for ttyfd: %s\n", ttyname(tty));
		for ( i=0; i<=2; ++i )
		{
			if ( isatty(i) )
				fprintf(stderr, "tty for descriptor: %d: %s\n",
						i, ttyname(i));
			else
				fprintf(stderr, "descriptor %d not a tty.\n", i);
		}
#endif /* DEBUG */

		unlink(holdfile);
		return(tty);
	}
	
	/* The parent exits so the child can be in the background. */

	fprintf(stderr, "\r\nDisconnecting from terminal.\r\n");
	die();
}
#endif /* SIGTSTP */

#ifdef HOST_NOT_FOUND
/* This is an error function similar to perror().
   Thanks Mike. :)    */

extern int h_errno;

int m_herror(str)
char *str;
{

  	printf("%s: ",str);
  	switch(h_errno)
  	{
    		case 0: 	     printf("No error.\n"); 
				     break;
    		case HOST_NOT_FOUND: printf("Host not found.\n"); 
				     break;
    		case TRY_AGAIN:      printf("Try again.\n"); 
				     break;
    		case NO_RECOVERY:    printf("No recovery.\n"); 
				     break;
    		case NO_DATA: 	     printf("No data or no addresss.\n"); 
			      	     break;
    		default: 	     printf("Unknown error.\n");
  	}
}
#endif /* HOST_NOT_FOUND */
