
/*   daemonize.c   Run any program as a daemon 

	This program will turn any program specified into
   a daemon.  The program will be run with a minimal environment
   ( HOME=/ and PATH=/bin:/usr/bin )  and will not be attatched 
   to any control terminal.  The program can re-attatch itself 
   to a new control terminal by opening up the terminal device
   it wants to attatch to.

	Shareware Copyright   	-Sam Lantinga	6/1/93
*/

#define	PATH	"/bin:/usr/bin:/etc"	/* Define this as the default $PATH */

#include	<sys/types.h>
#include	<sys/stat.h>
#include	<stdio.h>

extern char *getenv();
extern void  dropctty();		/* From misc.c  */
static char *psearch();


main(argc, argv)
int argc;
char *argv[];
{
	extern char **environ;		/* The global environment */

	int   pid;			/* The child process */
	int   fd;			/* Temporary filedes */
	char  path[1024];		/* The $PATH environment */
	char *execpath;			/* The command to execute */


	/* Check for a program to run */
	if ( ! argv[1] )
	{
		fprintf(stderr, "Usage: %s program [arg...]\n", argv[0]);
		exit(1);
	}
	else
		++argv;

	/* Check to see if we can run the program */
	for ( fd=1, execpath=argv[0]; *execpath; ++execpath )
	{
		if ( *execpath == '/' )
		{  /* We are a full pathname */
			fd=0;
			break;
		}
	}
	execpath=argv[0];

	if ( fd && ((execpath=psearch(execpath)) == NULL))
	{
		fprintf(stderr, "Can't find %s\n", argv[0]);
		exit(1);
	}
	

	/* Fork the child daemon process */
	if ( (pid=fork()) < 0 )
	{
		perror("Can't fork()");
		exit(2);
	}
	else if ( pid > 0 )
	{
		printf("Daemon running with pid: %d\n", pid);
		exit(0);
	}
	dropctty();		/* Disassociate from our tty */

	/* Set up the limited environment */
	
	sprintf(path, "PATH=%s", PATH);
	*environ="HOME=/";
	*(environ+1)=path;
	*(environ+2)=(char *)0;

	/* Now run the program on the command line */
	if ( execvp(execpath, argv) < 0 )
	{
		perror("exec() failed");
		exit(2);
	}
}


/* A function to search the user's path for an executable file
   and return the full pathname of that file. */

static char	exec_path[1024];	/* The actual pathname */
	
static char *psearch(executable)
char *executable;
{
	char *tptr, *pptr, *path;
	struct stat sb;
	int i, j=0, k=0;


	if ( (pptr=getenv("PATH")) == NULL )
		return(NULL);

	for ( tptr=pptr; *pptr; ++pptr )
	{
		if ( *pptr == ':' )
		{
			if ( tptr == pptr )	/* blah::/blah entry */
				path=executable;
			else
			{
				(*pptr)='\0';
				sprintf(exec_path, "%s/%s", tptr, executable);
				path=exec_path;
				(*pptr)=':';
			}

			if ( stat(path, &sb) < 0 )
				tptr=(pptr+1);
			else
				return(path);
		}
	}
	return(NULL);
}
